#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <asm/scatterlist.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/system.h>

#ifndef LINUX_VERSION_CODE
#include <linux/version.h>
#endif /* LINUX_VERSION_CODE */
#include <linux/iobuf.h>
#include "sgat_dio.h"

///////////////////////////////////////////////////////////////////////////////
//
//Defines
//
//////////////////////////////////////////////////////////////////////////////
#define DPRINTK(x...) printk("%s:",__FUNCTION__); printk(x)
#define FUNC(x...)  DPRINTK(x)
#define DPRINTK(x...)

#define SZ_SGAT_SCH sizeof(Sgat_dio_scatter_hold)
#define SZ_SGAT_KIOBUF_VALIST sizeof(struct kiobuf_valist)

struct kiobuf_valist
{
	unsigned long va;
	int		offset;		/* Offset to start of valid data */
	int		pgindex;
	int		pgcount;
	size_t	len;
};

typedef struct sgat_dio_scatter_hold  /* holding area for scsi scatter gather info */
{
    unsigned short k_use_sg;    /* Count of kernel scatter-gather pieces */
    unsigned dxfer_len;         /* Size of (aggregate) data buffer */
    void * buffer;              /* kernel scatter list*/
    struct kiobuf * kiobp;      /* for direct IO information */
    char mapped;                /* indicates kiobp has locked pages */
    void *kiobuf_valist;
}Sgat_dio_scatter_hold;    

///////////////////////////////////////////////////////////////////////////////
//
//Forward Declarations
//
//////////////////////////////////////////////////////////////////////////////
static int sgat_dio_build_iohdr(const Sgat_dio_hdr_t *io_hdrp, Sgat_dio_hdr_t * hp);
#if 0
static int sgat_dio_get_dxfer_len(struct iovec *iovecp,int iovec_count);
#endif
static void sgat_dio_finish_req(Sgat_dio_scatter_hold *schp);
static int sgat_dio_build_sgatl(Sgat_dio_scatter_hold * schp,int,
							    unsigned short);
static char * sgat_dio_malloc(int rqSz);
static void sgat_dio_free(char * buff);
static inline int sgat_dio_ms_to_jif(unsigned int msecs);
static inline unsigned sgat_dio_jif_to_ms(int jifs);
static int sgat_dio_map_user_kiobuf(int rw, struct kiobuf *iobuf,
			 void *valist, int limit);
static void force_phy_pages_allocation(unsigned long p,unsigned long o,int pgcount);
static __kernel_size_t  compute_tablesize(struct iovec *sglp,unsigned short iovec_count);

///////////////////////////////////////////////////////////////////////////////
//
//Utility Functions
//
//////////////////////////////////////////////////////////////////////////////
static inline int sgat_dio_ms_to_jif(unsigned int msecs)
{
    if ((UINT_MAX / 2U) < msecs)
	return INT_MAX;      /* special case, set largest possible */
    else
	return ((int)msecs < (INT_MAX / 1000)) ? (((int)msecs * HZ) / 1000)
					       : (((int)msecs / 1000) * HZ);
}

static inline unsigned sgat_dio_jif_to_ms(int jifs)
{
    if (jifs <= 0)
	return 0U;
    else {
	unsigned int j = (unsigned int)jifs;
	return (j < (UINT_MAX / 1000)) ? ((j * 1000) / HZ) : ((j / HZ) * 1000);
    }
}

static char * sgat_dio_malloc(int rqSz)
{
    char * resp = NULL;

    if (rqSz <= 0)
        return resp;
    resp = kmalloc(rqSz, GFP_KERNEL);
    if (resp) 
	    memset(resp, 0, rqSz);
    DPRINTK("size=%d,ret=0x%p\n", rqSz, resp);
    return resp;
}
static void sgat_dio_free(char * buff)
{
    FUNC("\n");
    if (! buff) return;
    DPRINTK("Free'ing %p\n",buff);
    kfree(buff);  
}
Sgat_dio_request *alloc_sgat_dio_request(void)
{
    FUNC("\n");
   return ((Sgat_dio_request *) sgat_dio_malloc(SZ_SGAT_REQ)); 
}
void free_sgat_dio_request(Sgat_dio_request *srp)
{
    FUNC("\n");
    sgat_dio_free((char *)srp );
}
static void sgat_dio_free_schp(Sgat_dio_scatter_hold *schp)
{
    FUNC("\n");
    if ( schp && schp->kiobuf_valist )
	sgat_dio_free((char *)schp->kiobuf_valist);
    /*Remove scatter list*/
    if ( schp && schp->buffer )
	sgat_dio_free((char *)schp->buffer);
    /*Free req*/
    if ( schp )
	sgat_dio_free((char *)schp);
}
static Sgat_dio_scatter_hold * sgat_dio_alloc_schp_and(int elems)
{
    FUNC("\n");
    Sgat_dio_scatter_hold *schp = NULL;
    schp = (Sgat_dio_scatter_hold *) sgat_dio_malloc(SZ_SGAT_SCH);    
    if ( !schp )
	return NULL;
    schp->kiobuf_valist = sgat_dio_malloc(elems*SZ_SGAT_KIOBUF_VALIST);
    if ( !schp->kiobuf_valist )
    {
	sgat_dio_free_schp(schp);
	return NULL;
    }
return schp;
}
static void sgat_dio_finish_req(Sgat_dio_scatter_hold * schp)
{
    FUNC("\n");
    if (schp && schp->kiobp) {
	if (schp->mapped) {
	    unmap_kiobuf(schp->kiobp);
	    DPRINTK("unmapping user pages\n");
	    schp->mapped = 0;
	}
	free_kiovec(1, &schp->kiobp);
	schp->kiobp = NULL;
	sgat_dio_free_schp(schp);
    }
}
static void force_phy_pages_allocation(unsigned long p,unsigned long o,int pgcount)
{
    int i;
    char x=0;

    //__copy_to_user(&x,(char *)p,1);
    FUNC("\n");
    put_user(x,(char *)p);
    for ( p += PAGE_SIZE-o,i = 1; i < pgcount; p += PAGE_SIZE,i++)
    {
	put_user(x,(char *)p);
	//__copy_to_user(&x,(char *)p,1);
    }
}
static int sgat_dio_map_user_kiobuf(int rw, struct kiobuf *iobuf, 
			 void *kiobuf_valist, int limit)
{
	int pgcount=0, err;
	int total_pages=0, total_len=0;
	int count,pg_index=0;
	struct mm_struct *	mm;
	struct kiobuf_valist *valist = kiobuf_valist;	
	/* Make sure the iobuf is not already mapped somewhere. */
	if (iobuf->nr_pages)
		return -EINVAL;

	mm = current->mm;
	DPRINTK ("begin\n");
	for ( count = 0; count < limit; count++,valist++)
	{
		if ( !valist->va )
			return -EFAULT;
		pgcount = (valist->va + valist->len + PAGE_SIZE - 1)/PAGE_SIZE - 
			    valist->va/PAGE_SIZE;
		/* mapping 0 bytes is not permitted */
		if (!pgcount) BUG();
		valist->offset = valist->va & (PAGE_SIZE-1);
		valist->pgcount = pgcount;
		total_pages +=pgcount;
		total_len +=valist->len;
		/*Bring all physical pages in*/
		if ( rw == READ ) 
		    force_phy_pages_allocation(valist->va,
					   valist->offset,
					   pgcount); 	    
	}
	DPRINTK("total pages %d\n",total_pages);	
	err = expand_kiobuf(iobuf, total_pages);
	if (err)
		return err;
	iobuf->locked = 0;
	iobuf->length = total_len;
	
	DPRINTK("entries %d\n",limit);
	/* Try to fault in all of the necessary pages */
	down_read(&mm->mmap_sem);
	valist = kiobuf_valist;	
	for ( count = 0; count < limit ; count ++ ,valist++)
	{
	    pgcount = (valist->va + valist->len + PAGE_SIZE - 1)/PAGE_SIZE 
			    - valist->va/PAGE_SIZE;
	    valist->pgindex = pg_index;
	    DPRINTK("pgindex %d pgcount %d\n",pg_index,pgcount);
	    /* rw==READ means read from disk, write into memory area */
	    err = get_user_pages(current, mm, valist->va, pgcount,
		    (rw==READ), 1, &iobuf->maplist[pg_index], NULL);
	    if (err < 0) {
		    unmap_kiobuf(iobuf);
		    DPRINTK ("end %d\n", err);
		    iobuf->nr_pages = 0;
		    up_read(&mm->mmap_sem);
		    return err;
	    }
	    pg_index +=pgcount;
	}
	iobuf->nr_pages = pg_index;
	up_read(&mm->mmap_sem);
	while (total_pages--) {
		/* FIXME: flush superflous for rw==READ,
		 * probably wrong function for rw==WRITE
		 */
	    flush_dcache_page(iobuf->maplist[total_pages]);
	}
	DPRINTK ("end OK\n");
	return 0;
}

static int sgat_dio_build_sgatl(Sgat_dio_scatter_hold * schp,
				int mx_sc_elems,
				unsigned short iovecc)
{
    int k,l;
    int  offset, rem_sz,useg_len;
    struct kiobuf * kp;
    struct scatterlist * sclp;
    int	pgindex =0;
    int pgcount=0;
    unsigned short iovec_count=0;
    struct kiobuf_valist *valist;

    /* scatter gather array, followed by mem_src_arr (array of chars) */
    int elem_sz = sizeof(struct scatterlist);

    schp->buffer = sgat_dio_malloc(mx_sc_elems * elem_sz);
    if (! schp->buffer)
	return -ENOMEM;

    memset(schp->buffer, 0, mx_sc_elems * elem_sz);

    kp = schp->kiobp;
    valist = schp->kiobuf_valist;
    for (k = 0,sclp = schp->buffer, rem_sz = schp->dxfer_len,iovec_count = 0;
	 (iovec_count < iovecc) && (rem_sz > 0 ) && (k < mx_sc_elems);
	 ++iovec_count,++valist ) {
	
	offset = valist->offset;
	useg_len = valist->len;
	pgcount = valist->pgcount;
	pgindex = valist->pgindex;
	DPRINTK ("iovec %d offset %d len"
	         "%d pgcount %d pgindex %d rem_sz %d\n",
		iovec_count,offset,useg_len,pgcount,pgindex,rem_sz);
        unsigned int len = useg_len;

	for ( l = pgindex; 
	    (l < pgindex+pgcount)&& (useg_len > 0) && (k < mx_sc_elems) ; 
	    ++l) {
	
	    if (l != pgindex ) {
		sclp->address = page_address(kp->maplist[l]);
		if ( l != (pgindex+pgcount-1)) {
		    sclp->length = PAGE_SIZE;
                } else {
		    sclp->length = useg_len;
                }
	    }
	    else{
		sclp->address = page_address(kp->maplist[l]) + offset;
                if (1 != pgcount) {
                    sclp->length = PAGE_SIZE - offset;
                }else {
                    sclp->length = useg_len;
                }

	    }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,13)
	    sclp->page = NULL;
#endif
	
	    DPRINTK ("offset %d len %d\n", offset,sclp->length);
            offset = 0;
	    useg_len -= sclp->length;
	    DPRINTK("k=%d, a=0x%p, len=%d useg_len = %d\n",
		k, sclp->address, sclp->length, useg_len);
	    ++k;
	    ++sclp;
	}
	rem_sz -= len;
	DPRINTK("rem_sz=%d\n",rem_sz);
    }
    if (rem_sz > 0) {   /* must have failed */
	return -EFAULT; 
    }
    schp->k_use_sg = k;
    DPRINTK("k_use_sg=%d, rem_sz=%d\n",k, rem_sz);
    return 0;
}

///////////////////////////////////////////////////////////////////////////////
//
//
//
//////////////////////////////////////////////////////////////////////////////
static __kernel_size_t compute_tablesize(struct iovec *sglp,unsigned short iovec_count)
{
    struct iovec *iovecp;
    int	i;
    __kernel_size_t scl_entries,sg_tablesize = 0;
    
    if (!sglp)
       return -EINVAL;	
    for (i= 0,iovecp = sglp;i < iovec_count ; i++,iovecp++)
    {
	scl_entries = (iovecp->iov_len / PAGE_SIZE) * 2 + 1;
	if ( 1 == scl_entries ) {
            scl_entries = 2 ;
        }
	sg_tablesize += scl_entries;
    }
    return sg_tablesize;
}
static int sgat_dio_build_iohdr(const Sgat_dio_hdr_t *io_hdrp, Sgat_dio_hdr_t * hp)
{
    int k;
    memset(hp,0,SZ_SGAT_DIO_HDR);

    FUNC("\n");
    if ((k = verify_area(VERIFY_READ, io_hdrp, SZ_SGAT_DIO_HDR)))
    {
	DPRINTK("failed verify_area\n");
	return k;  /* protects following copy_from_user()s + get_user()s */
    }
    __copy_from_user(hp, io_hdrp, SZ_SGAT_DIO_HDR);

    return 0;
}
int sgat_dio_build_sg_list(Sgat_dio_hdr_t *u_io_hdrp,
			    Sgat_dio_request *srp)
{

    int k,res,l;
    struct iovec *iovecp = NULL ;
    Sgat_dio_hdr_t *hp;
    Sgat_dio_scatter_hold *schp = NULL;
    int sg_tablesize;
    unsigned char *p;
    struct kiobuf_valist *valist;
   
    if ( !srp || !u_io_hdrp)
	return -EINVAL;

    hp = (Sgat_dio_hdr_t *) sgat_dio_malloc(SZ_SGAT_DIO_HDR);

    if ( !hp){
	DPRINTK("sgat_dio_malloc failed\n");
	return -ENOMEM;
    }

   if ( (k = sgat_dio_build_iohdr(u_io_hdrp,hp)))  
   {
       sgat_dio_free((char *)hp);
       return k;
   }

   srp->k_io_hdrp = hp;
   srp->u_io_hdrp = u_io_hdrp;
   srp->timeout = sgat_dio_ms_to_jif(hp->timeout);
   srp->dxfer_dir = hp->dxfer_dir;        /* [i] data transfer dir  */
   srp->xferId = hp->xferId;

    if ((hp->dxfer_dir != SGAT_DIO_DXFER_TO_DEV) && 
	(hp->dxfer_dir != SGAT_DIO_DXFER_FROM_DEV))
    {
	DPRINTK("unknown direction %d\n",hp->dxfer_dir);
	sgat_dio_free((char *)hp);
	return -EINVAL;
    }

    if ( (hp->iovec_count) && (hp->dxfer_len > 0) ) 
    {
	if ((k = verify_area(VERIFY_READ, hp->sglp,
			     SZ_SGAT_IOVEC * hp->iovec_count)))
	{
	    DPRINTK("failed verify_area\n");
	    sgat_dio_free((char *)hp);
	    return k;
	}
    }
    else
    {
	DPRINTK("invalid inputs(iovec count %d dxfer len %d)\n",
		hp->iovec_count,hp->dxfer_len);
	sgat_dio_free((char *)hp);
	return -EINVAL;
    }
    DPRINTK("dxfer_len %d\n",hp->dxfer_len);

    iovecp = (struct iovec *) sgat_dio_malloc(SZ_SGAT_IOVEC * hp->iovec_count);
    if ( !iovecp ){
	DPRINTK("sgat_dio_malloc failed\n");
	sgat_dio_free((char *)hp);
	return -ENOMEM;
    }
    schp = sgat_dio_alloc_schp_and(hp->iovec_count);
    if ( !schp || !schp->kiobuf_valist ){
	DPRINTK("sgat_dio_malloc failed\n");
	sgat_dio_finish_req(schp);
	sgat_dio_free((char *)iovecp);
	sgat_dio_free((char *)hp);
	return -ENOMEM;
    }
    __copy_from_user(iovecp, hp->sglp, SZ_SGAT_IOVEC * hp->iovec_count);

    for ( l =0, valist = schp->kiobuf_valist ; l < hp->iovec_count; l++,++valist)
    {
	p = (unsigned char *) iovecp[l].iov_base;
	if ((k = verify_area((SGAT_DIO_DXFER_TO_DEV == hp->dxfer_dir) ? 
			VERIFY_READ: VERIFY_WRITE, p, iovecp[l].iov_len)))
	{
	    sgat_dio_finish_req(schp);
	    sgat_dio_free((char *)iovecp);
	    sgat_dio_free((char *)hp);
	    return k;
	}
        DPRINTK("iovec num %d len %d \n",l,iovecp[l].iov_len);
	valist->va = (unsigned long)p;
	valist->len = iovecp[l].iov_len;
    }
    /*
     *Done with gathering user space iovecs' information
     */ 
    sgat_dio_free((char *)iovecp);

    res = alloc_kiovec(1, &schp->kiobp);
    if (0 != res) {
	DPRINTK("alloc_kiovec res=%d\n", res);
	sgat_dio_free((char *)hp);
	return -ENOMEM;
    }
    res = sgat_dio_map_user_kiobuf((SGAT_DIO_DXFER_TO_DEV == hp->dxfer_dir) ? 1 : 0,
			  schp->kiobp, schp->kiobuf_valist, hp->iovec_count);
    if (0 != res) {
	    DPRINTK("map_user_kiobuf res=%d\n", res);
	sgat_dio_finish_req(schp);
	sgat_dio_free((char *)hp);
	return -EFAULT;
    }
	
    /*
      map_user_kiobuf() does not seem to be guaranteeing consistency
      of the user and kernel mappings.  Flush and invalidate the
      data cache here, to make sure that the mappings are consistent
      with one another and with physical RAM.  dplatt@tivo.com 3/19/03
     */
    flush_dcache();
    schp->mapped =1;
    schp->dxfer_len = hp->dxfer_len;

    sg_tablesize = compute_tablesize(hp->sglp,hp->iovec_count);
    DPRINTK("sg_tablesize %d\n",sg_tablesize);
    if (sg_tablesize <= 0 ) {
	DPRINTK("Invalid sg_tablesize %d\n",sg_tablesize);
	sgat_dio_finish_req((Sgat_dio_scatter_hold *)schp);
	sgat_dio_free((char *)hp);
	return -EINVAL;
    }
    res = sgat_dio_build_sgatl(schp, sg_tablesize,hp->iovec_count);
    if (res < 0) {
	sgat_dio_finish_req((Sgat_dio_scatter_hold *)schp);
	sgat_dio_free((char *)hp);
	return res;
    }
    
    srp->sclp_count = schp->k_use_sg;    /* Count of kernel scatter-gather pieces */
    srp->sclp = schp->buffer;
    srp->schp = (void *) schp;
    return 0;
}
void sgat_dio_done( Sgat_dio_request *srp)
{
    FUNC("\n");
    if (!srp )
	return;	
    if ( srp->k_io_hdrp )
    {
        srp->k_io_hdrp->status = srp->status;
        srp->k_io_hdrp->client_data = srp->client_data;
	if (srp->u_io_hdrp )
	    copy_to_user(srp->u_io_hdrp, srp->k_io_hdrp, SZ_SGAT_DIO_HDR);
	sgat_dio_free((char *)srp->k_io_hdrp);
    }
    sgat_dio_finish_req((Sgat_dio_scatter_hold *)srp->schp);
}

void sgat_dio_done_nocopy( Sgat_dio_request *srp)
{
    FUNC("\n");
    if (!srp )
	return;	
    if ( srp->k_io_hdrp )
    {
        srp->k_io_hdrp->status = srp->status;
        srp->k_io_hdrp->client_data = srp->client_data;
	sgat_dio_free((char *)srp->k_io_hdrp);
    }
    sgat_dio_finish_req((Sgat_dio_scatter_hold *)srp->schp);
}
///////////////////////////////////////////////////////////////////////////////
