//////////////////////////////////////////////////////////////////////
//
// File: smblite.c
//
// Description: A light weight client wrapper around the Samba
// client code.  XXX: not very light weight at the moment...
//
// Some bits based on smbclient and nmblookup code
// Copyright (C) Andrew Tridgell 1994-1998
//
// Modified to accept simple commands via a control pipe
// Copyright (c) 2002 TiVo Inc.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//////////////////////////////////////////////////////////////////////

#include <includes.h>

#define DEBUG_SMBLITE

enum DataTransportType {
    DATA_TRANSPORT_UNKNOWN = 0,
    DATA_TRANSPORT_PIPE = 1,
    DATA_TRANSPORT_PIPE_ALIGNED = 2,
    // ...
};

enum DataDirection {
    DATA_DIRECTION_UNKNOWN = 0,
    DATA_DIRECTION_OUT = 1,
    DATA_DIRECTION_IN = 2,
    DATA_DIRECTION_BOTH = 3,
};

enum FileMode {
    FILE_MODE_NONE = 0,
    FILE_MODE_READ = 1,
    FILE_MODE_WRITE = 2,
    // ...
};

struct ConnectionStatus {
    struct cli_state state;
    int fConnected;
};

struct FileStatus {
    int handle;
    offset_t offset;   // current read/write offset
    struct ConnectionStatus* pConnection;   // NULL for not opened
};

struct ClientStatus {
    struct ConnectionStatus connection[2];
    struct FileStatus file[2];
    struct ConnectionStatus* pCurrentConnection;
    int nFiles;
    struct FileStatus* pCurrentFile;
    char* pBuffer;
    size_t sBuffer;
    offset_t offset;    // current buffer offset
    size_t sBlock;
    enum DataTransportType type;
    enum DataDirection direction;
    int dataFd;
    int fTerminate;
    uint16 hCurrentCommand;
};

static void
DoAckFirst( struct ClientStatus* pClient, const char* code )
{
    int lenAddl = strlen( code ) - 1;

    fprintf( stdout, "%c%04hx", code[0], pClient->hCurrentCommand );
    if ( lenAddl > 0 ) {
        if ( fwrite( &code[1], sizeof( char ), lenAddl, stdout ) != lenAddl ) {
            pClient->fTerminate = 1;
        }
    }
}

static void
DoAck( struct ClientStatus* pClient, const char* code )
{
    DoAckFirst( pClient, code );
    fflush( stdout );
}

static void
DoAckString( const char* string )
{
    fprintf( stdout, "%04x%s", strlen( string ), string );
}

static void
DoAckStringLast( const char* string )
{
    DoAckString( string );
    fflush( stdout );
}

int
DoOpenFile( struct ClientStatus* pClient, char* filename, int mode )
{
    int hFile;
    struct FileStatus* pFile;
    struct FileStatus* pOtherFile;
    struct ConnectionStatus* pConnection = pClient->pCurrentConnection;
    struct cli_state* cli = &pConnection->state;

    if ( pConnection->fConnected == 0 ) {
        return -1;
    }

    if ( pClient->file[0].pConnection == NULL ) {
        pFile = &pClient->file[0];
        pOtherFile = &pClient->file[1];
    } else if ( pClient->file[1].pConnection == NULL ) {
        pFile = &pClient->file[1];
        pOtherFile = &pClient->file[0];
    } else {
#ifdef DEBUG_SMBLITE
        fprintf( stderr, "Exceeded max 2 files...\n" );
#endif
        return -1;
    }

    hFile = cli_open( cli, filename, mode, DENY_NONE );
    if ( hFile == -1 ) {
        return -1;
    }

    pFile->handle = hFile;
    pFile->offset = 0;
    pFile->pConnection = pConnection;

    // set this as current file only if it's the only file open,
    // othrwise wait for the other file to close
    //
    if ( pOtherFile->pConnection == NULL ) {
        pClient->pCurrentFile = pFile;
    }

    return 0;
}

int
DoCloseFile( struct ClientStatus* pClient )
{
    BOOL s;
    struct FileStatus* pFile = pClient->pCurrentFile;

    if ( pFile->pConnection == NULL ) {
        return -1;
    }

    s = cli_close( &pFile->pConnection->state, pFile->handle );
    // always mark the file closed, even if cli_close() failed.
    pFile->pConnection = NULL;

    if ( pFile == &pClient->file[0] ) {
        pClient->pCurrentFile = &pClient->file[1];
    } else {
        pClient->pCurrentFile = &pClient->file[0];
    }

    return s ? 0 : -1;
}

int
DoStatCurrent( struct ClientStatus* pClient )
{
    struct FileStatus* pFile = pClient->pCurrentFile;
    struct cli_state* cli;
    uint16 attr;
    size_t size;
    time_t cTime;
    time_t aTime;
    time_t mTime;
    char ackBuf[38];

    if ( pFile->pConnection == NULL ) {
        return -1;
    }

    cli = &pFile->pConnection->state;
    if ( !cli_qfileinfo( cli, pFile->handle, &attr, &size, &cTime, &aTime,
                         &mTime, NULL, NULL ) &&
         !cli_getattrE( cli, pFile->handle, &attr, &size, &cTime, &aTime,
                        &mTime ) )
    {
        return -1;
    }
    // Ack S[U16:attributes][U32:size][U32 creation time][U32 access time]
    // [U32 modification time]: file stats
    snprintf( ackBuf, 38, "S%04hx%08x%08lx%08lx%08lx", attr &
              (int)( aRONLY | aHIDDEN | aSYSTEM | aVOLID | aDIR | aARCH ),
              size, cTime, aTime, mTime );
    ackBuf[37] = 0;
    DoAck( pClient, ackBuf );
    return 0;
}

// For now, this doesn't actually cancel anything, since we process
// commands serially.  When we read ahead commands, it will do something
// useful
//
void
DoCancelPending( struct ClientStatus* pClient )
{
    char ackBuf[6];

    // Ack P[U16:handle of first cancelled command]: pending ops canceled
    snprintf( ackBuf, 6, "P%04hx", pClient->hCurrentCommand );
    ackBuf[5] = 0;
    DoAck( pClient, ackBuf );
}

static void
BrowseFunc(const char *name, uint32 m, const char *comment, void *state)
{
    struct ClientStatus* pClient = (struct ClientStatus*)state; 

    // only support disk shared for now
    //
    if ( m == STYPE_DISKTREE ) {
        // Ack H[string:share name][string:share comment]: available share
        // on the connected host
        DoAckFirst( pClient, "H" );
        DoAckString( name );
        DoAckStringLast( comment );
    }
}

int
DoBrowseHost( struct ClientStatus* pClient )
{
    struct ConnectionStatus* pConnection = pClient->pCurrentConnection;
    struct cli_state* cli = &pConnection->state;

    if ( pConnection->fConnected == 0 ) {
        return -1;
    }

    if ( !cli_RNetShareEnum( cli, BrowseFunc, pClient ) ) {
        return -1;
    }
    return 0;
}

static void
ListFunc( file_info* finfo, const char* mask, void* state )
{
    struct ClientStatus* pClient = (struct ClientStatus*)state; 
    char ackBuf[38];

    // Ack I[U16:attributes][U32:size][U32 creation time][U32 access time]
    // [U32 modification time][string:filename without path]: file listing
    snprintf( ackBuf, 38, "I%04hx%08lx%08lx%08lx%08lx", finfo->mode &
              (int)( aRONLY | aHIDDEN | aSYSTEM | aVOLID | aDIR | aARCH ),
              finfo->size, finfo->ctime, finfo->atime, finfo->mtime );
    ackBuf[37] = 0;
    DoAckFirst( pClient, ackBuf );
    DoAckStringLast( finfo->name );
}

int
DoListFiles( struct ClientStatus* pClient, char* mask )
{
    struct ConnectionStatus* pConnection = pClient->pCurrentConnection;
    struct cli_state* cli = &pConnection->state;
    uint16 attribute = aDIR | aSYSTEM | aHIDDEN;

    if ( pConnection->fConnected == 0 ) {
        return -1;
    }

    if ( !cli_list( cli, mask, attribute, ListFunc, pClient ) ) {
        return -1;
    }
    return 0;
}

static int
DoOpenBroadcastSocket( )
{
    int fd;

    fd = open_socket_in( SOCK_DGRAM, 137, 0, 0, True );

    // if port 137 fails, try unpriveledged port.  this is mostly
    // for testing on x86
    //
    if ( fd == -1 ) {
        fd = open_socket_in( SOCK_DGRAM, 0, 3, 0, True );
    }

    if ( fd != -1 ) {
        set_socket_options( fd, "SO_BROADCAST" );
    }

    return fd;
}

int
DoQueryHosts( struct ClientStatus* pClient )
{
    int j, count;
    struct in_addr *ip_list=NULL;
    int fd;

    const char* lookup = "*";
    int lookup_type = 0;
    struct in_addr *bcast;

    fd = DoOpenBroadcastSocket();
    if (fd == -1) {
        return -1;
    }

    for ( j=iface_count() - 1; !ip_list && j >= 0 ; j-- ) {
        bcast = iface_n_bcast( j );
        ip_list = name_query( fd, lookup, lookup_type, True, True, *bcast,
                              &count );
    }

    if (!ip_list) {
        close( fd );
        return -1;
    }

    for ( j=0; j < count; j++ ) {
        char ackBuf[10];

        // Ack Q[U32:ip address]: browsable host
        snprintf( ackBuf, 10, "Q%08x", ntohl( ip_list[j].s_addr ) );
        ackBuf[9] = 0;
        DoAck( pClient, ackBuf );
    }

    free( ip_list );

    close( fd );
    return 0;
}

int
DoEnumerateServices( struct ClientStatus* pClient, struct in_addr* ip )
{
    struct nmb_name nname;
    int count, i;
    struct node_status *status;
    int fd;

    fd = DoOpenBroadcastSocket();
    if (fd == -1) {
        return -1;
    }

    make_nmb_name( &nname, "*", 0 );
    status = node_status_query( fd, &nname, *ip, &count );
    if ( status != NULL ) {
        for ( i=0; i < count ; i++ ) {
            if ( status[i].type == 0x20 && ( status[i].flags & 0x06 ) != 0 ) {
                // Ack T[string:service]: enumerated service
                DoAckFirst( pClient, "T" );
                DoAckStringLast( status[i].name );
            }
        }
        free( status );
    }

    close( fd );
    return 0;
}

// note: assumes file already checked for openness
//
static ssize_t
DoReadBlock( struct ClientStatus* pClient, size_t size )
{
    char* pData;
    struct FileStatus* pFile = pClient->pCurrentFile;
    struct cli_state* cli = &pFile->pConnection->state;
    ssize_t nRead, nWritten;
    char ackBuf[10];
    size_t sRead, nWrittenTot;

    if ( size > pClient->sBlock ) {
        sRead = pClient->sBlock;
    } else {
        sRead = size;
    }

    if ( pClient->offset + sRead > pClient->sBuffer ) {
        // this never happens for DATA_TRANSPORT_PIPE_ALIGNED transport
        // type, since offset always increments in sBlock sized chunks,
        // and sBuffer is a multiple of sBlock
        sRead = pClient->sBuffer - pClient->offset;
    }
    pData = pClient->pBuffer + pClient->offset;

    nRead = cli_read( cli, pFile->handle, pData, pFile->offset, sRead );
    if ( nRead > 0 ) {
        if ( pClient->offset == 0 ) {
            // Ack U[U32:size]: read data, start of buffer
            // Ack V[U32:size]: read data, continuing in buffer
            snprintf( ackBuf, 10, "U%08x", nRead );
        } else {
            snprintf( ackBuf, 10, "V%08x", nRead );
        }
        ackBuf[9] = 0;
        DoAck( pClient, ackBuf );

        pFile->offset += nRead;
        if ( pClient->type == DATA_TRANSPORT_PIPE_ALIGNED ) {
            pClient->offset += pClient->sBlock;
        } else {
            pClient->offset += nRead;
        }
        if ( pClient->offset >= pClient->sBuffer ) {
            pClient->offset = 0;
        }

        // XXX: this is the data out interface, just piping to FD 3
        // right now...
        //
        nWrittenTot = 0;
        while ( nWrittenTot < nRead ) {
            nWritten = write( pClient->dataFd , pData, nRead );
            if ( nWritten < 0 && errno != EINTR ) {
#ifdef DEBUG_SMBLITE
                fprintf( stderr, "Writing data out...\n" );
#endif
                return -1;
            }
            if ( nWritten > 0 ) {
                nWrittenTot += nWritten;
            }
        }
    }
    return nRead;
}

int
DoRead( struct ClientStatus* pClient, size_t size )
{
    struct FileStatus* pFile = pClient->pCurrentFile;
    ssize_t nReadTot;

    if ( pFile->pConnection == NULL || size == 0 || pClient->pBuffer == NULL ) {
        return -1;
    }

    nReadTot = 0;
    while ( nReadTot < size ) {
        ssize_t nRead;
        size_t sRead = size - nReadTot;

        nRead = DoReadBlock( pClient, sRead );
        if ( nRead < 0 && nReadTot == 0 ) {
            return -1;
        } else if ( nRead <= 0 ) {
            break;
        }
        nReadTot += nRead;
    }

    // Ack W: read done with less data than requested
    if ( nReadTot < size ) {
        DoAck( pClient, "W" );
    }

    return 0;
}

int
DoSeek( struct ClientStatus* pClient, off_t offset )
{
    struct FileStatus* pFile = pClient->pCurrentFile;
    if ( pFile->pConnection == NULL ) {
        return -1;
    }

    pFile->offset = offset;

    return 0;
}

// XXX: need to fix read so that we read socket directly into this buffer...
int
DoSetBuffer( struct ClientStatus* pClient, size_t sBuffer, size_t sBlock,
             enum DataTransportType type, enum DataDirection direction )
{
    if ( sBuffer == 0 || sBlock == 0 || sBlock > sBuffer ) {
        return -1;
    }

    // only read buffer supported right now
    if ( direction != DATA_DIRECTION_OUT ) {
        return -1;
    }

    switch ( type ) {
        // for now, only support pipe transport
      case DATA_TRANSPORT_PIPE_ALIGNED:
        if ( sBuffer % sBlock != 0 ) {
            return -1;
        }
        // fall through
      case DATA_TRANSPORT_PIPE:
        if ( pClient->pBuffer == NULL || pClient->sBuffer != sBuffer ) {
            if ( pClient->pBuffer != NULL ) {
                free( pClient->pBuffer );
            }
            // since this is just our side of the buffer right now, we
            // don't have to worry about aligning.
            pClient->pBuffer = malloc( sBuffer );
        }
        if ( pClient->pBuffer == NULL ) {
            return -1;
        }
        break;

      default:
        return -1;
    }

    pClient->sBuffer = sBuffer;
    pClient->offset = 0;
    pClient->sBlock = sBlock;
    pClient->type = type;
    pClient->direction = direction;

    return 0;
}

int
DoConnect( struct ClientStatus* pClient, const char* server,
           const char* sharename, char* username, char* password,
           char* workgroup, struct in_addr* ip )
{
    struct cli_state *c;
    struct nmb_name called, calling;
    const char* pSharename;
    char *p;
    struct ConnectionStatus* pConnection;

    if ( sharename[0] == 0 ) {
        pSharename = "IPC$";
    } else {
        pSharename = sharename;
    }

    if ( pClient->connection[0].fConnected == 0 ) {
        pConnection = &pClient->connection[0];
    } else if ( pClient->connection[1].fConnected == 0 ) {
        pConnection = &pClient->connection[1];
    } else {
#ifdef DEBUG_SMBLITE
        fprintf( stderr, "Exceeded max 2 connections...\n" );
#endif
        return -1;
    }

    c = &pConnection->state;

    make_nmb_name(&calling, "TiVo", 0x0);
    make_nmb_name(&called , server, 0x20);

    while (1) {
        /* have to open a new connection */
        if (!(c=cli_initialise(c)) || (cli_set_port(c, SMB_PORT) == 0) ||
            !cli_connect(c, server, ip)) {
            return -1;
        }

        c->protocol = PROTOCOL_NT1;

        if (cli_session_request(c, &calling, &called)) {
            break;
        }

        cli_shutdown(c);

        if ((p=strchr(called.name, '.'))) {
            *p = 0;
            continue;
        }
        if (strcmp(called.name, "*SMBSERVER")) {
            make_nmb_name(&called , "*SMBSERVER", 0x20);
            continue;
        }
        return -1;
    }

    if (!cli_negprot(c)) {
        cli_shutdown(c);
        return -1;
    }

    if (!cli_session_setup(c, username, 
                           password, strlen(password),
                           password, strlen(password),
                           workgroup)) {
        /* if a password was not supplied then try again with a null username */
        if (password[0] || !username[0] || 
            !cli_session_setup(c, "", "", 0, "", 0, workgroup)) { 
            cli_shutdown(c);
            return -1;
        }
    }

#if 0 // we should probably log this somewhere...
    if (*c->server_domain || *c->server_os || *c->server_type)
            DEBUG(1,("Domain=[%s] OS=[%s] Server=[%s]\n",
                    c->server_domain,c->server_os,c->server_type));
#endif

    if (!cli_send_tconX(c, pSharename, "?????", password, strlen(password)+1)) {
        cli_shutdown(c);
        return -1;
    }

    pConnection->fConnected = 1;
    // note that we flip the current connection over to this one even
    // if we're still connected to the other one, so disconnect cannot
    // just close current connection
    pClient->pCurrentConnection = pConnection;

    return 0;
}

int
DoDisconnect( struct ClientStatus* pClient )
{
    struct ConnectionStatus* pConnection;
    struct ConnectionStatus* pOtherConnection;

    // if both connections are open, we want to close the non-current
    // connection
    pConnection = pClient->pCurrentConnection;
    if ( pConnection == &pClient->connection[0] ) {
        pOtherConnection = &pClient->connection[1];
    } else {
        pOtherConnection = &pClient->connection[0];
    }
    if ( pOtherConnection->fConnected != 0 ) {
        pConnection = pOtherConnection;
    }

    if ( pConnection->fConnected == 0 ) {
        return -1;
    }

    // make sure any files that were open on this connection get
    // closed
    //
    while ( pClient->file[0].pConnection == pConnection ||
            pClient->file[1].pConnection == pConnection ) {
        DoCloseFile( pClient );
    }

    cli_shutdown( &pConnection->state );
    pConnection->fConnected = 0;
    // don't need to flip pCurrentConneciton here, since that is done
    // at DoConnect time

    return 0;
}

// doesn't seem like dataFd really belongs here, but it is comes from
// the commandline args, so until we have a data transport that doesn't
// need it, I'm not going to worry about where else it ought to be...
//
void
DoInitClient( struct ClientStatus* pClient, int dataFd )
{
    memset( pClient, 0, sizeof( struct ClientStatus ) );

    pClient->pCurrentConnection = &pClient->connection[0];
    pClient->pCurrentFile = &pClient->file[0];
    pClient->dataFd = dataFd;
}

void
DoCleanupClient( struct ClientStatus* pClient )
{
    while ( pClient->pCurrentFile->pConnection != NULL ) {
        DoCloseFile( pClient );
    }

    while ( pClient->pCurrentConnection->fConnected != 0 ) {
        DoDisconnect( pClient );
    }
}

static int GetU16( uint16* pData )
{
    char dataBuf[4];

    if ( fread( dataBuf, sizeof( char ), 4, stdin ) != 4 ||
         sscanf( dataBuf, "%4hx", pData ) != 1 ) {
        return -1;
    }
    return 0;
}

static int GetU32( uint32* pData )
{
    char dataBuf[8];

    if ( fread( dataBuf, sizeof( char ), 8, stdin ) != 8 ||
         sscanf( dataBuf, "%8x", pData ) != 1 ) {
        return -1;
    }
    return 0;
}

static char* GetString( )
{
    char* pString;
    uint16 sString;

    if ( GetU16( &sString ) != 0 ) {
        return NULL;
    }

    pString = malloc( sString + 1 );
    if ( pString == NULL ) {
        return NULL;
    }

    if ( sString != 0 ) {
        if ( fread( pString, sizeof( char ), sString, stdin ) != sString ) {
            free ( pString );
            return NULL;
        }
    }

    pString[ sString ] = 0;

    return pString;
}

static void
ProcessOneCommand( struct ClientStatus *pClient )
{
    int s = 0;
    int nRead;
    char cmdBuf[1];

    do {
        nRead = fread( cmdBuf, sizeof( char ), 1, stdin );
        if ( nRead <= 0 ) {
            pClient->fTerminate = 1;
            return;
        }
    // swallow end-of-command newlines to assist debugging
    } while ( cmdBuf[0] == '\n' );

    if ( GetU16( &pClient->hCurrentCommand ) != 0 ) {
        pClient->fTerminate = 1;
        return;
    }

    switch ( cmdBuf[0] ) {
        // Cmd N[U32:IP address][string:server Netbios name][string:share name]
        // [string:username][string:password][string:workgroup]: Open a
        // connection to a server and share and make it the current connection
      case 'N':
        {
            char* pServer;
            char* pShare;
            char* pUsername;
            char* pPassword;
            char* pWorkgroup;
            struct in_addr dest_ip;
            uint32 tempAddr;

            if ( GetU32( &tempAddr ) == 0 &&
                 ( pServer = GetString() ) != NULL &&
                 ( pShare = GetString() ) != NULL &&
                 ( pUsername = GetString() ) != NULL &&
                 ( pPassword = GetString() ) != NULL &&
                 ( pWorkgroup = GetString() ) != NULL ) {
                dest_ip.s_addr = htonl( tempAddr );
                s = DoConnect( pClient, pServer, pShare, pUsername, pPassword,
                               pWorkgroup, &dest_ip );
                free( pServer );
                free( pShare );
                free( pUsername );
                free( pPassword );
                free( pWorkgroup );
            } else {
                s = -1;
                pClient->fTerminate = 1;
            }
        }
        break;

        // Cmd M: Close the least recent connection
      case 'M':
        s = DoDisconnect( pClient );
        break;

        // Cmd O[U16 mode][string:filename]: Open a file on the current
        // connection. Note that it will not be made the current file
        // until after the previously opened file is closed, if 2 files
        // are opened at once.
      case 'O':
        {
            uint16 temp;
            enum FileMode mode;
            char* pFilename;

            if ( GetU16( &temp ) == 0 && ( pFilename = GetString() ) != NULL ) {
                mode = (enum FileMode)temp;
                // insist on read-only, nothing else supported right now
                if ( mode == FILE_MODE_READ ) {
                    s = DoOpenFile( pClient, pFilename, O_RDONLY );
                } else {
                    s = -1;
                }
                free( pFilename );
            } else {
                s = -1;
                pClient->fTerminate = 1;
            }
        }
        break;

        // Cmd L: Close least recently opened file
      case 'L':
        s = DoCloseFile( pClient );
        break;

        // Cmd K[U32: position]: Set file read/write offset
      case 'K':
        {
            uint32 position;

            if ( GetU32( &position ) != 0 ) {
                s = -1;
                pClient->fTerminate = 1;
                break;
            }
            s = DoSeek( pClient, position );
        }
        break;

        // Cmd R[U32 size]: Read bytes from current open file and
        // increment file read/write offset by bytes read
      case 'R':
        {
            uint32 size;

            if ( GetU32( &size ) != 0 ) {
                s = -1;
                pClient->fTerminate = 1;
                break;
            }
            s = DoRead( pClient, size );
            if ( s == 0 ) {
                return;
            }
        }
        break;

        // Cmd U[U32:buffer size][U32:block size][U16:transport type]
        // [U16:data direction]:
        // Set buffer params. For optimal usage, buffer size should be a
        // multiple of block size, and reads should done in multiples of
        // block size when possible. Buffer size should be as small as
        // possible, block size should be as large as possible (these
        // need to be balanced per user's needs).
      case 'U':
        {
            uint32 sBuffer, sBlock;
            uint16 temp1, temp2;
            enum DataTransportType type;
            enum DataDirection direction;

            if ( GetU32( &sBuffer ) != 0 || GetU32( &sBlock ) != 0 ||
                 GetU16( &temp1 ) != 0 || GetU16( &temp2 ) ) {
                s = -1;
                pClient->fTerminate = 1;
                break;
            }
            type = (enum DataTransportType)temp1;
            direction = (enum DataDirection)temp2;
            s = DoSetBuffer( pClient, sBuffer, sBlock, type, direction );
        }
        break;

        // Cmd S: Get file stats on the current open file
      case 'S':
        s = DoStatCurrent( pClient );
        if ( s == 0 ) {
            return;
        }
        break;

        // Cmd P: Cancel pending queued commands
      case 'P':
        DoCancelPending( pClient );
        return;

        // Cmd H: Enumerate browsable shares on the connected host
      case 'H':
        s = DoBrowseHost( pClient );
        break;

        // Cmd I[string:mask]: List files matching a pattern on the
        // connected share. Use \path\* to get a directory listing of
        // the \path\ directory
      case 'I':
        {
            char* pMask;

            if ( ( pMask = GetString() ) != NULL ) {
                s = DoListFiles( pClient, pMask );
                free( pMask );
            } else {
                s = -1;
                pClient->fTerminate = 1;
            }
        }
        break;

        // Cmd T[U32:IP address]: Enumerate services on a host
      case 'T':
        {
            struct in_addr dest_ip;
            uint32 tempAddr;

            if ( GetU32( &tempAddr ) != 0 ) {
                s = -1;
                pClient->fTerminate = 1;
                break;
            }

            dest_ip.s_addr = htonl( tempAddr );
            s = DoEnumerateServices( pClient, &dest_ip );
        }
        break;

        // Cmd Q: query hosts on the local subnet
      case 'Q':
        s = DoQueryHosts( pClient );
        break;

        // Cmd X: exit
      case 'X':
        pClient->fTerminate = 1;
        return;

      default:
        s = -1;
        pClient->fTerminate = 1;
        break;
    }

    if ( s == 0 ) {
        // Ack K: command completed status OK
        DoAck( pClient, "K" );
    } else {
        // Ack R[U16:error code]: command completed with error
        // XXX: error code not meaningful right now...
        DoAck( pClient, "R0000" );
    }
}

static void ProcessCommands( int dataFd )
{
    struct ClientStatus client;

    DoInitClient( &client, dataFd );

    while ( client.fTerminate == 0 ) {
        ProcessOneCommand( &client );
    }

    DoCleanupClient( &client );

    return;
}


int main ( int argc, char **argv )
{
    int dataFd;

    if ( argc != 2 || ( dataFd = atoi( argv[1] ) ) <= 2 ) {
        fprintf( stderr, "TiVo Samba client started (based on "
                         "Samba version %s).\n", VERSION );
        fprintf( stderr, "Usage: %s dataFd\n", argv[0] );
        return 1;
    }

    // XXX: need to review all this...
    TimeInit();
    charset_initialise();
#if 1 // XXX: this seems unsafe to not do, so leaving it in for now...
    {
        extern BOOL in_client;
        in_client = True;
    }
    lp_load("",True,False,False);
    codepage_initialise(lp_client_code_page());
#else
    codepage_initialise( MSDOS_LATIN_1_CODEPAGE );
#endif
    load_interfaces();

    ProcessCommands( dataFd );

    return 0;
}
